#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Xiang Wang <ramwin@qq.com>


from django.urls import path

from . import views


urlpatterns = [
        path("call-command/", views.CallCommandView.as_view()),
]



from django.conf import settings

from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework.views import APIView

from .tasks import async_call_command


class CallCommandSerializer(Serializer):
    using = serializers.ChoiceField(
            choices=["thread", "local", "celery"],
            required=False, default="local")
    command = serializers.CharField()
    args = serializers.ListField(required=False, default=list)
    kwargs = serializers.DictField(required=False, default=dict)

    class Meta:
        fields = ["async", "command", "args", "kwargs"]


class CallCommandView(APIView):

    def post(self, request, *args, **kwargs):
        serializer = CallCommandSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = serializer.validated_data
        command = data["command"]
        if command not in settings.DJANGO_COMMANDS_ALLOW_REMOTE_CALL:
            raise PermissionDenied(f"you are not allowed to call command `{command}`")
        if data["using"] == "celery":
            async_call_command.delay(
                    data["command"],
                    data["using"],
                    data["args"],
                    data["kwargs"],
            )
        else:
            async_call_command(
                    data["command"],
                    data["using"],
                    data["args"],
                    data["kwargs"],
            )
        return Response({})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Xiang Wang <ramwin@qq.com>


import logging
import time

from threading import Thread
from django.test import TestCase
from django.core.management import call_command
from django_commands.management.commands.test_wait_commands import Command as TestWaitCommand


handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
LOGGER = logging.getLogger("django_commands")
LOGGER.addHandler(handler)
LOGGER.setLevel(logging.DEBUG)


class Test(TestCase):

    def test_wait_command(self):
        TestWaitCommand.create_task(1)
        TestWaitCommand.create_task(2)
        TestWaitCommand.create_task(3)
        thread = Thread(group=None, target=call_command, args=["test_wait_commands"])
        thread.start()
        time.sleep(0.1)
        print("创建任务")
        for i in range(4, 20):
            TestWaitCommand.create_task(i)



import logging
import time

from django.test import TestCase
from django_commands import models, utils
from django_commands.models import CommandLog
from django_commands.tasks import async_call_command
from django_commands.utils import get_middle_string, iter_large_queryset

from rest_framework.test import APIClient

LOGGER = logging.getLogger(__name__)


class BisectTask(utils.Bisect):
    DATA = [0, 1, 2, 3, 4, 5, 6, 8, 9, 10]

    def has_error(self, start, end):
        return (self.DATA[end] - self.DATA[start]) != end - start


class TestAsyncCommand(TestCase):

    def test_async(self):
        start = time.time()
        async_call_command("slow_command", using="thread")
        async_call_command("slow_command", using="thread")
        LOGGER.info("you may want run `celery -A project worker` to check if the delay is working")
        async_call_command.delay("slow_command", using="celery")
        async_call_command.delay("slow_command", using="celery")
        self.assertGreater(
                start + 1,
                time.time(),
        )

    def test_api(self):
        start = time.time()
        client = APIClient()
        res = client.post("/api/django-commands/call-command/",
                          {"command": "slow_command", "using": "thread"}, format="json")
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json(), {})

        self.assertGreater(
                start + 1,
                time.time(),
        )

        res = client.post("/api/django-commands/call-command/",
                          {"command": "slow_command"}, format="json")
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json(), {})

        self.assertGreater(
                start + 6,
                time.time(),
        )

    def test_deny_command(self):
        client = APIClient()
        res = client.post("/api/django-commands/call-command/",
                          {"command": "non_exist_command"}, format="json")
        self.assertEqual(res.status_code, 403)


class TestUtil(TestCase):

    def test(self):
        for i in range(1000):
            CommandLog.objects.create(
                    name=str(i)
            )
        iterator = iter_large_queryset(
                CommandLog.objects.filter(name__endswith="1"),
                batch_size=10)
        self.assertEqual(
            len(list(iterator)),
            10
        )
        results = []
        for queryset in iter_large_queryset(
                CommandLog.objects.filter(name__endswith="1")):
            results.extend([command.name for command in queryset])
        self.assertEqual(len(queryset), 100)

    def test_bisect(self):
        self.assertAlmostEqual(
            BisectTask(0, 9, 1).find_first_error(),
            6.5,
            delta=0.5)

    def test_get_middle_string(self):
        self.assertEqual(get_middle_string("a", "c"), "b")
        self.assertEqual(get_middle_string("", "c"), " ")
        self.assertEqual(get_middle_string("ba", "bc"), "bb")
        self.assertEqual(get_middle_string("", ""), "")
        self.assertEqual(get_middle_string("", "2"), " ")
        self.assertEqual(get_middle_string("7125811f-67c5-47f6-8327-b570b6ce72a1", "7136ec17-5a50-4d4c-a643-f3c1506bfeab"), "712a")

    def test_none_large_queryset(self):
        self.assertEqual(
            [i for i in iter_large_queryset(CommandLog.objects.all())],
            []
        )

    def test_iter_large_queryset(self):
        CommandLog.objects.create(name="bbb")
        CommandLog.objects.create(name="ccc")
        CommandLog.objects.create(name="aaa")
        a, b = iter_large_queryset(CommandLog.objects.all(), batch_size=2, ordering_field="name")
        self.assertEqual(
            b.get().name, "ccc")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Xiang Wang <ramwin@qq.com>


"""
datetime option
"""


import datetime
import itertools
import re

from typing import Iterable

from django.db.models import QuerySet, Q, ForeignKey
from django.utils import timezone


TZ = timezone.get_default_timezone()


def datetime_type(datetimestr):
    """
    It will convert three kind of format:
        monthdaystring:
            03-22 >> 2024-03-22T00:00:00+defaulttimezone
        datestring:
            2024-03-22 >> 2024-03-22T00:00:00+defaulttimezone
        datetimestring:
            "2024-03-22 01:02:03" >> 2024-03-22T01:02:03+defaulttimezone
        isodatetimestring:
            "2024-03-22T01:02:03+04:00" >> 2024-03-22T01:02:03+04:00

    Usage:

    def add_arguments(self, parser):
        parser.add_argument("datetime", type=datetime_type)
    """
    if re.match(r"^\d{+}-\d{+}$", datetimestr):
        month, day = datetimestr.split("-")
        return datetime.datetime(
                datetime.datetime.now().year, int(month), int(day),
                0, 0, 0,
                tzinfo=TZ,
        )
    if re.match(r"^\d{4}-\d{2}-\d{2}$", datetimestr):
        year, month, day = datetimestr.split("-")
        return datetime.datetime(
                int(year), int(month), int(day),
                0, 0, 0,
                tzinfo=TZ,
        )
    datetime_obj = datetime.datetime.fromisoformat(datetimestr)
    if timezone.is_aware(datetime_obj):
        return datetime_obj
    return timezone.make_aware(datetime_obj)


def iter_large_queryset(queryset, batch_size: int = 256, ordering_field: str = "pk") -> Iterable[QuerySet]:
    """
    split queryset in batch_size, so you can use multiprocess to handler it.
    mechanism:
    1. find the offset batch_size object, get iss id
    2. use id to filter queryset, get the next batch_size offset

    so the queryset must be order by pk/id
    """
    queryset = queryset.order_by(ordering_field)
    first_obj = queryset.first()
    if first_obj is None:
        return
    start_value = getattr(first_obj, ordering_field)
    while True:
        try:
            end_value = getattr(
                    queryset.all()[batch_size],
                    ordering_field,
            )
        except IndexError:
            if queryset.exists():
                yield queryset
            return
        q = Q(**{
            f"{ordering_field}__gte": start_value,
            f"{ordering_field}__lt": end_value,
        })
        result = queryset.filter(q)
        if result.exists():
            yield result
        else:
            return
        start_value = end_value
        queryset = queryset.filter(**{f"{ordering_field}__gte": start_value})


class Bisect:

    def __init__(self, start, end, step, auto_check=True):
        """find first error in duration"""
        self.start = start
        self.end = end
        self.step = step
        self._checked = False
        if auto_check:
            self.check()

    def check(self):
        if not self.has_error(self.start, self.end):
            raise ValueError("no error between start and end")
        self._checked = True

    def has_error(self, start, end):
        raise NotImplementedError

    def find_first_error(self):
        if not self._checked:
            raise ValueError("call check before find the first error")
        start = self.start
        end = self.end
        while (end - start) > self.step:
            middle = self.get_middle(start, end)
            if self.has_error(start, middle):
                end = middle
            else:
                start = middle
        return start

    def get_middle(self, start, end):
        return (start + end) // 2 


def get_middle_string(lower: str, upper: str) -> str:
    """
    return a middle value between lower and upper, in case you need to do bisect task for database field like uuid.
    """
    result = ""
    for lower_letter, upper_letter in itertools.zip_longest(lower, upper):
        if lower_letter is None:
            if ord(upper_letter) >= 32:
                return result + " "
            return result
        if lower_letter == upper_letter:
            result += lower_letter
            continue
        if ord(upper_letter) - ord(lower_letter) == 1:
            return result + lower_letter + 'a'
        return result + chr(
                (ord(upper_letter) + ord(lower_letter)) // 2
        )
    return result


class Dependency:

    def __init__(self, models, max_count=100):
        self.graph = TopologicalSorter()
        self.objects = {None}
        self.pending = set(models)
        self.max_count = max_count

    def update_objects(self) -> None:
        while self.pending and len(self.objects) <= self.max_count:
            model = self.pending.pop()
            self.objects.add(model)
            self.update_object(model)

    def update_object(self, model) -> None:
        for field in model._meta.fields:
            if not isinstance(field, ForeignKey):
                continue
            dependency = getattr(model, field.name)
            if dependency in self.objects:
                continue
            self.pending.add(dependency)
            self.graph.add(model, dependency)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Xiang Wang <ramwin@qq.com>


"""
Usefule Command Class
"""


import datetime
import logging
import time

from contextvars import ContextVar
from decimal import Decimal
from multiprocessing import Pool
from typing import Tuple, Union, Iterable, Optional

from redis import Redis

from django.core.management.base import BaseCommand, CommandParser
from django.db import connections
from django.utils import timezone

from django_redis import get_redis_connection

import django_commands

from .mixins import AutoLogMixin, WarmShutdownMixin
from .utils import iter_large_queryset


PROCESS_INITED = ContextVar("inited", default=False)
LOGGER = logging.getLogger(__name__)


class AutoLogCommand(BaseCommand):
    """
    AutoLogCommand will add log to every exception and then raise it
    """

    def execute(self, *args, **kwargs):
        try:
            return super().execute(*args, **kwargs)
        except Exception as error:
            LOGGER.exception(error)
            raise

    def handle(self, *args, **kwargs):
        raise NotImplementedError


class UniqueCommand(AutoLogCommand):
    """
    UniqueCommand can assert that only one command instance is running

    How it works?
    Every time the command executes:

        1. it will create a django_commands.models.CommandLog instance
        2. it will check if there is another Command Instance pending with the same UNIQUE_NAME

    """
    UNIQUE_NAME = ""
    TIMEOUT = datetime.timedelta(days=1)

    def execute(self, *args, **kwargs):
        unique_name = self.get_unique_name()
        new_instance = django_commands.models.CommandLog.objects.create(name=unique_name)
        wait_after = timezone.now() - self.TIMEOUT
        exist_commands = django_commands.models.CommandLog.objects.filter(
                status="pending",
                name=unique_name,
                create_datetime__gte=wait_after
        ).exclude(pk=new_instance.pk)
        if exist_commands.exists():
            new_instance.status = "skipped"
            new_instance.save()
            return
        try:
            self.handle()
        finally:
            new_instance.status = "finished"
            new_instance.save()

    def get_unique_name(self) -> str:
        """
        get a unique name for this command
        """
        if self.UNIQUE_NAME:
            return self.UNIQUE_NAME
        return f"{self.__class__.__module__}.{self.__class__.__name__}"

    def handle(self, *args, **kwargs):
        raise NotImplementedError


class WaitCommand(AutoLogMixin, BaseCommand):
    """
    A WaitCommand will use the redis blopop
    to run a command as soon as possible

    before start it will delete the key, so you can
    call create_task many times and only one time will it execute

    class variable:
        SQUASH_TASK = True  # wheter squash multi tasks into one
    """
    NAME = ""
    IMMEDIATELY = False
    SQUASH_TASK = True
    BATCH_SIZE = 1  # how many tasks do you want to pop once
    MAX_TIMES = 0  # how many tasks do you want to handle approximately, 0 means unlimit

    def add_arguments(self, parser: CommandParser):
        parser.add_argument(
                "--times", type=int, default=0,
                help="how many times do you want to execute at most")
        super().add_arguments(parser)

    def handle(self, *args, **kwargs) -> None:
        """
        get a unique name for this command
        """
        self.before_handle()
        if self.IMMEDIATELY:
            self.create_task()
        redis, redis_key = self.get_redis_info()
        LOGGER.info("rpush %s to trigger task", redis_key)
        max_run_time = kwargs.get("times", 0)
        run_time = 0
        while True:
            if self.MAX_TIMES and run_time >= self.MAX_TIMES:
                LOGGER.info("reach the max times limit, stop command: %s", self)
                break
            if max_run_time and run_time >= max_run_time:
                LOGGER.info("%s has executed as least %d times, bye bye", self, run_time)
                return
            key_taskids = redis.lpop(redis_key, self.BATCH_SIZE)
            if key_taskids is None:
                key_taskid = redis.blpop(redis_key, timeout=5)
                if key_taskid is None:
                    LOGGER.debug("no task")
                    continue
                LOGGER.debug("get one task: %s", key_taskid)
                key_taskids = [key_taskid[1]]
            else:
                LOGGER.debug("get multi task one time: %s", key_taskids)
            if self.SQUASH_TASK:
                redis.delete(redis_key)
            LOGGER.debug("handle task: %s", key_taskids)
            for key_taskid in key_taskids:
                self.handle_task(key_taskid, *args, **kwargs)
                run_time += 1

    def before_handle(self) -> None:
        """
        before handle hook
        this will be called only once
        """
        return

    def handle_task(self, task_id, *args, **kwargs) -> None:
        """
        override the handle_task function to do the real task
        """
        LOGGER.error("new task created, you should override this function")

    @classmethod
    def get_redis_info(cls) -> Tuple[Redis, str]:
        redis = get_redis_connection("default")
        redis_key = cls.NAME or f"{cls.__module__}.{cls.__name__}"
        return redis, redis_key

    @classmethod
    def create_task(cls, task_id: Optional[str] = None) -> None:
        redis, key = cls.get_redis_info()
        task_id = task_id or str(time.time())
        redis.rpush(key, task_id)


class MultiTimesCommand(AutoLogMixin, WarmShutdownMixin, BaseCommand):
    """
    MultiTimesCommand will run multi times according to INTERVAL AND MAX_TIMES

    you can use `kill -TERM <processid>` to kill the command
    you can set MAX_TIMES to decimal.Decimal("inf") to run forever
    """
    INTERVAL = 1
    MAX_TIMES: Union[Decimal, int] = 60
    run_cnt = 0

    def execute(self, *args, **kwargs):
        while self.run_cnt < self.MAX_TIMES:
            self.run_cnt += 1
            try:
                super().execute(*args, **kwargs)
            except StopIteration:
                break
            time.sleep(self.INTERVAL)


class RunForeverCommand(MultiTimesCommand):
    """
    Run Forever Command will run the commands forever. until you call

        kill -TERM <processid>

    """
    MAX_TIMES = Decimal("inf")


class DurationCommand(AutoLogCommand):
    """
    DurationCommand will run the command multi times until the running time exceed the MAX_DURATION(default 1 minute)
    """
    INTERVAL = 1
    DURATION = datetime.timedelta(minutes=1)

    def execute(self, *args, **kwargs):
        duration = self.DURATION
        end_time = datetime.datetime.now() + duration
        LOGGER.info("end_time: %s", end_time)
        while datetime.datetime.now() < end_time:
            try:
                self.handle(*args, **kwargs)
            except StopIteration:
                LOGGER.debug("StopIteration occured, DurationCommand will exit")
                break
            time.sleep(self.INTERVAL)

    def handle(self, *args, **kwargs):
        raise NotImplementedError


class MultiProcessCommand(AutoLogCommand):
    """
    A multiprocess command will use the multiprocessing.Pool to handler task.  
    You need to inherit it and custom the `get_tasks` and `handle_single_task`.  
    """

    def add_arguments(self, parser: CommandParser):
        parser.add_argument(
                "-j", "--jobs", type=int,
                help="how many process")
        super().add_arguments(parser)

    def handle(self, *args, jobs=None, **kwargs):
        start = timezone.now()
        tasks = self.get_tasks()
        LOGGER.debug("handle task: %s", tasks)
        with Pool(jobs) as p:
            for result in p.imap_unordered(
                self.handle_single_task, tasks):
                LOGGER.info("handle single task done: %s", result)

    @classmethod
    def handle_single_task(cls, *args, **kwargs):
        """handle single task"""
        if PROCESS_INITED.get() is False:
            PROCESS_INITED.set(True)
            # connections.close_all()
        raise NotImplementedError

    def get_tasks(self):
        """return iterable task"""
        raise NotImplementedError


class LargeQuerysetMutiProcessHandlerCommand(MultiProcessCommand):
    queryset = None
    DURATION = datetime.timedelta(minutes=1)
    BATCH_SIZE = 256
    MAX_TASK = Decimal("inf")

    def get_queryset(self):
        if self.queryset is None:
            raise ValueError("Please set queryset on the commands")
        return self.queryset

    def get_tasks(self) -> Iterable[Tuple[int, int]]:
        """use iter_large_queryset util to iterate a large queryset"""
        end_datetime = timezone.now() + self.DURATION
        results = []
        for queryset in iter_large_queryset(self.get_queryset(), self.BATCH_SIZE):
            if timezone.now() > end_datetime:
                break
            if not queryset:
                break
            assert queryset.first()
            assert queryset.last()
            results.append((queryset.first().pk, queryset.last().pk))
            if len(results) >= self.MAX_TASK:
                break
        connections.close_all()
        return results

    @classmethod
    def handle_single_task(cls, *args, **kwargs):
        """handle single task"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Xiang Wang <ramwin@qq.com>

# pylint: disable=too-few-public-methods

import logging
import signal


LOGGER = logging.getLogger(__name__)


class AutoLogMixin:

    def handle(self, *args, **kwargs):
        try:
            super().handle(*args, **kwargs)
        except Exception as error:
            LOGGER.exception(error)
            raise


class WarmShutdownMixin:
    """
    This Mixin enables your command to accept TERM signal and exist after current task finished 
    After inherit this mixin, you should add code like this:

        if self.need_stop:
            return

    in your command
    """
    need_stop = False

    def execute(self, *args, **kwargs):
        signal.signal(signal.SIGTERM, self.handle_signal)
        super().execute(*args, **kwargs)

    def handle_signal(self, signalnum, handler):  # pylint: disable=unused-argument
        LOGGER.info("receive stop signal")
        self.need_stop = True



# Generated by Django 4.2.7 on 2023-12-06 11:42

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CommandLog',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField()),
                ('status', models.TextField(choices=[('pending', 'pending'), ('skipped', 'skipped'), ('finished', 'finished')], default='pending')),
                ('create_datetime', models.DateTimeField(auto_now_add=True)),
                ('update_datetime', models.DateTimeField(auto_now=True)),
            ],
            options={
                'indexes': [models.Index(fields=['status', 'name'], name='django_comm_status_c6d84b_idx')],
            },
        ),
    ]



from django.db import models

# Create your models here.


class CommandLog(models.Model):

    STATUS_CHOICES = (
        ("pending", "pending"),
        ("skipped", "skipped"),
        ("finished", "finished"),
    )

    name = models.TextField()
    status = models.TextField(choices=STATUS_CHOICES, default="pending")
    create_datetime = models.DateTimeField(auto_now_add=True)
    update_datetime = models.DateTimeField(auto_now=True)

    class Meta:
        indexes = [
            models.Index(fields=["status", "name"])
        ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Xiang Wang <ramwin@qq.com>


import logging
from threading import Thread
from typing import Literal

from django.core.management import call_command
from celery import shared_task


LOGGER = logging.getLogger(__name__)


@shared_task
def async_call_command(
        command: str,
        using: Literal["celery", "thread", "local"]="thread",
        args=None, kwargs=None,
    ) -> None:
    args = args or []
    kwargs = kwargs or {}
    if using == "celery":
        LOGGER.info("celery task started")
        call_command(command, *args, **kwargs)
        return
    if using == "thread":
        LOGGER.info("thread task started")
        Thread(None, call_command, args=[command, *args], kwargs=kwargs).start()
        return
    if using == "local":
        LOGGER.info("local task started")
        call_command(command, *args, **kwargs)
        return
    raise NotImplementedError(using)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Xiang Wang <ramwin@qq.com>


from django.core.management.base import BaseCommand
from django_commands.utils import datetime_type


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument("datetime", nargs="+", type=datetime_type)

    def handle(self, *args, **kwargs):
        print(kwargs["datetime"])
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Xiang Wang <ramwin@qq.com>


import logging
from django_commands.commands import WaitCommand


LOGGER = logging.getLogger(__name__)


class Command(WaitCommand):
    SQUASH_TASK = False
    BATCH_SIZE = 2
    MAX_TIMES = 10

    def handle_task(self, taskid, **kwargs):
        LOGGER.info(taskid)



# SPDX-FileCopyrightText: 2023-present Xiang Wang <ramwin@qq.com>


"""
some useful command sub class
"""


import datetime
import logging
import time
from decimal import Decimal

from django.core.management.base import BaseCommand

from .commands import (
        MultiTimesCommand,
        RunForeverCommand,
        AutoLogCommand,
        DurationCommand,
        )


LOGGER = logging.getLogger(__name__)
